%{
/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2016  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: parser.y 273 2016-03-18 10:59:02Z cau $
 */    
#include "system.h"
void yyerror(char const *);
int yylex ( YYSTYPE * lvalp, YYLTYPE * llocp );
/*#define YYPARSE_PARAM scanner
  #define YYLEX_PARAM   scanner*/
%}
%locations
%pure-parser
%union {
    int tokentype;
    NODEPTR nodetype;
}

%debug

%start program

%type <nodetype> program non_empty_expression_end_list expression
%type <nodetype> expression_list non_empty_expression_list
%type <nodetype> non_empty_name_list name_list  for_range
%type <nodetype> list_expression list sublist list_expr set_expression
%type <nodetype> atom
%type <nodetype> error

%token <tokentype> MINTOKEN 	/* To define codetable size. */

/* internal tokens */
%token <tokentype> FORTIMES FORIN MAKELIST SUB SUBLIST
%token <tokentype> BOOL LOC STATE STATIC LIST CALL iCALL
%token <tokentype> iEXISTS iPROCESS iPREFIX iFOR  iFORTIMES iFORIN iPROJ
%token <tokentype> iSEMI iAND ENV BND MEMBLOCK iIF iSYSTEM iLEN
%token <tokentype> iOR iUNDEFINED iREMEMBER

/* regular tokens */
%nonassoc <tokentype> RUN
%nonassoc <tokentype> HELP
%token <tokentype> WHILE FOR IN TIMES REPEAT  
%token <tokentype> EXISTS EXISTSF FORALL
%token <tokentype> TO SIZE
%token <tokentype> TRUE FALSE LAMBDA SKIP EMPTY MORE RANLEN INFINITE END
%token  <nodetype> NAME STRING FLOAT INT
%left  <tokentype> PROJ
%token <tokentype> FIXLIST VARLIST NPREV SHOWMEM 
%token  <nodetype> EXIT SHOWTRACE SHOWSTATE ABORT BREAK RETURN CONTINUE
%token  <nodetype> LEAP STEP
%token  <nodetype> PASSLIMIT VERBOSITY PRINTLEVEL DEBUGLEVEL STATEPRINT
%token  <nodetype> REDUNDANCY CALLBYNAME ASSIGNAHEAD BREAKISABORT HOPCHOP
%token  <nodetype> PATH HELPDIR INFILE OUTFILE TIMESTEP
%token  <nodetype> RANDSEED MAXRANDLEN PRECISION
%token <tokentype> RANDOM  RRANDOM  SRANDOM  SRRANDOM STATENUMBER
%token <tokentype> NIL
%token <tokentype> LBRACE
%left  <tokentype> PROCESS
%left  <tokentype> LPAREN
%left  <tokentype> LBRACK
%token <tokentype> RBRACE
%token <tokentype> RPAREN
%token <tokentype> RBRACK
%token <tokentype> BAR
%left  <tokentype> SEMI
%left  <tokentype> IMPLIES
%left  <tokentype> OR
%left  <tokentype> WHERE
%left  <tokentype> AND
%token <tokentype> IF THEN
%nonassoc <tokentype> IFX
%nonassoc <tokentype> ELSE
%nonassoc <tokentype> DEFINE SET FORMAT  ATANTWO 
%nonassoc <tokentype> DO UNTIL HISTORY
%nonassoc <tokentype> GETS IS ASSN UASSN SKIPUASSN DEFAULTS
%left <tokentype> KEEP
%nonassoc <tokentype> THE LT LE EQ GE GT NEQ
%left  <tokentype> LEN
%left  <tokentype> PLUS  MINUS
%left  <tokentype> MULT DIV MOD
%left  <tokentype> POWER
%left  <tokentype> PREFIX ALWAYS SOMETIME HALT FIN INIT STB MEM 
%left  <tokentype> INPUT WINPUT OUTPUT PEEK  CHOPSTAR 
%left  <tokentype> NOT NEXT WNEXT PREV TYPE ASCII STRUCT LOAD SYSTEM
%left  <tokentype> UMINUS UPLUS CEIL  FLOOR SQRT ITOF EXPO LOG LOGTEN 
%left  <tokentype> SIN COS TAN ASIN ACOS ATAN SINH COSH TANH FABS
%left  <tokentype> REFER DEREF 
%token <tokentype> COMMA

%nonassoc <tokentype> MAXTOKEN	/* To define codetable size. */

/* tokens for error messages */
%token <tokentype> MQS IDN MDF EIF H1 STL IPTL ERA UNK

%expect 4

%% /* Rules start here */

program:
non_empty_expression_end_list { $$=$1; Parseroot=$$;  if (INTERACTIVE) return(0);}

non_empty_expression_end_list:
program expression END    { $$ = node2(AND,$1,$2);  }
| expression END  { $$=$1;   }
;

/* EXPRESSIONS */

expression:
  LBRACE expression RBRACE { $$ = $2; }
| LPAREN expression RPAREN { $$ = $2; }

/* CONDITIONALS */
| IF expression THEN expression %prec IFX { $$ = node3(IF,$2,$4,True); }
| IF expression THEN expression ELSE expression { $$ = node3(IF,$2,$4,$6); }
| expression IMPLIES expression { $$ = node2(IMPLIES,$1,$3); }

/* REPETITION */
| WHILE expression DO expression { $$ = node2(WHILE,$2,$4); }
| FOR NAME IN expression DO expression { $$ = node3(FORIN,$2,$4,$6); }
| FOR expression TIMES DO expression { $$ = node2(FORTIMES,$2,$5); }
| FOR for_range DO expression { $$ = node3(FOR,node_lnk_1($2),node_lnk_2($2),$4); }
| CHOPSTAR expression { $$ = node1(CHOPSTAR,$2); }
| REPEAT expression UNTIL expression { $$ = node2(REPEAT,$2,$4); }

/* QUANTIFICATION */
| EXISTS name_list DO expression { $$ = node5(EXISTS,$2,NULL,$4,One,False); }
| EXISTSF name_list DO expression { $$ = node5(EXISTS,$2,NULL,$4,One,True); }
| EXISTS name_list HISTORY expression DO expression {
    if (node_ival($4)<1 || node_type($4)!= INT)
            syntax_error(H1);
    $$ = node5(EXISTS,$2,NULL,$6,$4,False); }
| EXISTSF name_list HISTORY expression DO expression {
    if (node_ival($4)<1 || node_type($4)!= INT)
            syntax_error(H1);
    $$ = node5(EXISTS,$2,NULL,$6,$4,True); }
| EXISTS for_range DO expression { $$ = node5(EXISTS,node_lnk_1($2),node_lnk_2($2),$4,One,False); }
| EXISTS for_range HISTORY expression DO expression { if (node_ival($4)<1 || node_type($4)!= INT)
            syntax_error(H1);
      $$ = node5(EXISTS,node_lnk_1($2),node_lnk_2($2),$6,$4,False); }
| FORALL for_range DO expression { $$ = node3(FORALL,node_lnk_1($2),node_lnk_2($2),$4); }

/* FUNCTIONS */
| RUN expression { $$ = node2(RUN,$2,True); }
| LOAD expression { $$ = node1(LOAD,$2); }
| FORMAT LPAREN expression_list RPAREN { $$ = node1(FORMAT,$3); }
| OUTPUT expression { $$ = node1(OUTPUT,$2); }
| WINPUT expression { $$ = node1(WINPUT,$2); }
| INPUT expression { $$ = node1(INPUT,$2); }
| PEEK expression { $$ = node1(PEEK,$2); }
| DEFINE NAME EQ expression { $$ = node2(DEFINE,$2,$4); }
| DEFINE NAME LPAREN name_list RPAREN EQ expression { $$ = node2(DEFINE,$2,node2(LAMBDA,$4,$7)); }
| TYPE expression { $$ = node1(TYPE,$2); }
| ASCII expression { $$ = node1(ASCII,$2); }
| SET set_expression EQ expression { $$ = node2(SET,$2,$4); }
| HELP expression { $$ = node1(HELP,$2); }
| HELP { $$ = node1(HELP,NULL); }
| SYSTEM expression { $$ = node1(SYSTEM,$2); }
| NAME LPAREN expression_list RPAREN { $$ = node2(CALL,$1,$3); }
| LAMBDA LPAREN name_list RPAREN DO expression { $$ = node2(LAMBDA,$3,$6); }
| THE expression DO expression { $$ = node2(THE,$2,$4); }
| SHOWMEM LPAREN expression COMMA expression RPAREN  { $$=node2(SHOWMEM,$3,$5);    }

/* BOOLEAN OPERATORS */
| expression OR expression { $$ = node2(OR,$1,$3); }
| expression AND expression { $$ = node2(AND,$1,$3);  }
| expression WHERE expression { $$ = node2(WHERE,$1,$3); }
| NOT expression { $$ = node1(NOT,$2); }

/* TEMPORAL OPERATORS */
| expression SEMI expression { $$ = node2(SEMI,$1,$3); }
| expression GETS expression { $$ = node2(GETS,$1,$3); }
| expression IS expression { $$ = node2(IS,$1,$3); }
| expression ASSN expression { $$ = node2(ASSN,$1,$3); }
| expression UASSN expression { $$ = node2(UASSN,$1,$3); }
| expression SKIPUASSN expression { $$ = node2(SKIPUASSN,$1,$3); }
| expression DEFAULTS expression { $$ = node2(DEFAULTS,$1,$3); }
| PROCESS expression { $$ = node1(PROCESS,$2); }
| PREFIX expression { $$ = node1(PREFIX,$2); }
| expression PROJ expression { $$ = node2(PROJ,$1,$3); }
| ALWAYS expression { $$ = node1(ALWAYS,$2); }
| SOMETIME expression { $$ = node1(SOMETIME,$2); }
| HALT expression { $$ = node1(HALT,$2); }
| LEN expression { $$ = node1(LEN,$2); }
| MEM expression { $$ = node1(MEM,$2); }
| FIN expression { $$ = node1(FIN,$2); }
| INIT expression { $$ = node1(INIT,$2); }
| KEEP expression { $$ = node1(KEEP,$2); }
| STB expression { $$ = node1(STB,$2); }
| STRUCT expression { $$ = node1(STRUCT,$2); }
| NEXT expression { $$ = node1(NEXT,$2); }
| WNEXT expression { $$ = node1(WNEXT,$2); }
| PREV expression { $$ = node1(PREV,$2); }
| NPREV LPAREN expression COMMA expression RPAREN { $$ = node2(NPREV,$3, $5); }

/* RELATIONAL OPERATORS */
| expression LT expression { $$ = node2(LT,$1,$3); }
| expression LE expression { $$ = node2(LE,$1,$3); }
| expression EQ expression { $$ = node2(EQ,$1,$3); }
| expression GE expression { $$ = node2(GE,$1,$3); }
| expression GT expression { $$ = node2(GT,$1,$3); }
| expression NEQ expression { $$ = node2(NEQ,$1,$3); }

/* POINTER OPERATORS */
| REFER expression { $$ = node1(REFER,$2); }
| DEREF expression { $$ = node1(DEREF,$2); }

/* ARITHMETIC OPERATORS */
| PLUS expression %prec UPLUS { $$ = node1(UPLUS,$2); }
| MINUS expression %prec UMINUS { $$ = node1(UMINUS,$2); }
| expression PLUS expression { $$ = node2(PLUS,$1,$3); }
| expression MINUS expression { $$ = node2(MINUS,$1,$3); }
| expression MULT expression { $$ = node2(MULT,$1,$3); }
| expression DIV expression { $$ = node2(DIV,$1,$3); }
| expression MOD expression { $$ = node2(MOD,$1,$3); }
| expression POWER expression { $$ = node2(POWER,$1,$3); }
| CEIL expression { $$ = node1(CEIL,$2); }
| FLOOR expression { $$ = node1(FLOOR,$2); }
| SQRT expression { $$ = node1(SQRT,$2); }
| ITOF expression { $$ = node1(ITOF,$2); }
| EXPO expression { $$ = node1(EXPO,$2); }
| LOG expression { $$ = node1(LOG,$2); }
| LOGTEN expression { $$ = node1(LOGTEN,$2); }
| SIN expression { $$ = node1(SIN,$2); }
| COS expression { $$ = node1(COS,$2); }
| TAN expression { $$ = node1(TAN,$2); }
| ASIN expression { $$ = node1(ASIN,$2); }
| ACOS expression { $$ = node1(ACOS,$2); }
| ATAN expression { $$ = node1(ATAN,$2); }
| SINH expression { $$ = node1(SINH,$2); }
| COSH expression { $$ = node1(COSH,$2); }
| TANH expression { $$ = node1(TANH,$2); }
| FABS expression { $$ = node1(FABS,$2); }
| ATANTWO LPAREN expression COMMA expression RPAREN { $$ = node2(ATANTWO,$3,$5); }



/* LIST OPERATORS */
| FIXLIST LPAREN expression COMMA expression RPAREN { $$ = node2(FIXLIST,$3,$5); }
| VARLIST LPAREN expression COMMA expression RPAREN { $$ = node2(VARLIST,$3,$5); }
| BAR expression BAR %prec SIZE { $$ = node1(SIZE,$2); }

/* LISTS & ATOMS */
| list_expression
| atom
| error
;

list_expression :
  list
| sublist
;

list:
  LBRACK expression_list RBRACK { $$ = node1(MAKELIST,$2); }
;

sublist:
  list_expr LBRACK RBRACK { $$ = node2(SUB,$1,NULL); }
| list_expr LBRACK expression RBRACK { $$ = node2(SUB,$1,$3); }
| list_expr LBRACK expression TO expression RBRACK { $$ = node3(SUBLIST,$1,$3,$5); }
;

list_expr:
 NAME
| NAME LPAREN expression_list RPAREN  { $$ = node2(CALL,$1,$3); }
| list_expression
;




atom:
  NAME      
| TRUE      { $$=True;    }
| FALSE     { $$=False;    }
| NIL       { $$=Nil;    }
| STRING    
| EMPTY     { $$=Empty;    }
| SKIP      { $$=Skip;    }
| MORE      { $$=More;    }
| RANLEN    { $$=Ranlen;    }
| INFINITE  { $$=Infinite;    }
| STATENUMBER { $$=StateNr; }
| RANDOM    { $$=Random;    }
| RRANDOM   { $$=RRandom;    }
| SRANDOM   { $$=SRandom;    }
| SRRANDOM  { $$=SRRandom;    }
| LEAP      
| STEP      
| BREAK     
| CONTINUE  
| RETURN    
| ABORT     
| SHOWTRACE 
| SHOWSTATE 
| EXIT
| INT       
| FLOAT
;

set_expression:
  PASSLIMIT
| VERBOSITY
| PRINTLEVEL
| DEBUGLEVEL
| STATEPRINT
| REDUNDANCY
| CALLBYNAME
| ASSIGNAHEAD
| BREAKISABORT
| HOPCHOP
| PATH
| HELPDIR
| INFILE
| OUTFILE
| TIMESTEP
| RANDSEED
| MAXRANDLEN
| PRECISION
;

name_list :
  /* empty */ { $$ = NULL; }
| non_empty_name_list
;

non_empty_name_list:
  NAME
| NAME COMMA non_empty_name_list { $$ = node2(COMMA,$1,$3); }
;                                     
                                     
expression_list:
  /* empty */ { $$ = NULL; }
| non_empty_expression_list
| LPAREN expression_list RPAREN { $$ = $2; }
;

non_empty_expression_list:
expression { $$ = node2(COMMA,$1,NULL); }
| expression COMMA non_empty_expression_list { $$ = node2(COMMA,$1,$3); }
;

for_range:
NAME LT expression { $$ = node2(LT,$1,$3); }
| LPAREN NAME LT expression RPAREN  { $$ = node2(LT,$2,$4); }   
;



%%

NODEPTR parse(stream)
FILE *stream;
{
    NODEPTR p;

    if (InputFile != stream) {
        EOFILE=0;
        if (yydebug)
            printf("restarting parse\n");
        yyrestart(stream);
        InputFile = stream;
    }
        
    yydebug = (DebugLevel>=6) ? 1 : 0;
    
    yyerrorcount = 0;
        
    if ((yyparse () == 0) && (yyerrorcount == 0)) {
        if (yydebug) {
             printf("in parse-old type before\n");
             output_expn(Parseroot,-1);
             printf("\n after \n");
        }
        p= resolve_refs (Parseroot);
        if (yydebug) {
            output_expn(p,-1);
            printf("\n");
            printf("EOFILE=%d\n",EOFILE);
        }
        return(p);
    }
    else {
        if (yydebug)
            printf("yyerrorcount %d\n", yyerrorcount);
        return(True);
    }
        
}

NODEPTR resolve_refs (e)
NODEPTR e;
{
    if (e == NULL || e == LastCommand)
        return (e);
    switch (node_type(e)) {
        case NAME:
	    return (lookup_env (node_sym_name(e)));
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LOC:
            return (e);
        case EXISTS:
            push_scope2 (node_lnk_1(e),node_ival(node_lnk_4(e)),node_bval(node_lnk_5(e))? 1 : 0);
	    node_lnk_3(e) = resolve_refs (node_lnk_3(e));
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case FORALL:
        case FORIN:
        case FOR:
	    push_scope2 (node_lnk_1(e),1,0);
	    node_lnk_3(e) = resolve_refs (node_lnk_3(e));
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case LAMBDA:
	    push_scope2 (node_lnk_1(e),1,0);
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case IF:
        case SUBLIST:
            node_lnk_3(e) = resolve_refs (node_lnk_3(e));
        case THE:
        case FORTIMES:        
        case PROJ:
        case COMMA:
        case DEFINE:
        case WHILE:
        case REPEAT:
        case LT:
        case LE:
        case EQ:
        case GE:
        case GT:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case SEMI:
        case WHERE:
        case AND:
        case OR:
        case IMPLIES:    
        case GETS:
        case IS:
        case MOD:
        case ATANTWO:
        case ASSN:
        case UASSN:
        case SKIPUASSN:
        case CALL:
        case SUB:
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case RUN:
        case SET:
        case SHOWMEM:
        case DEFAULTS:
            node_lnk_2(e) = resolve_refs (node_lnk_2(e));
        case NOT:
        case UPLUS:
        case UMINUS:
        case CEIL:
        case FLOOR:
        case SQRT:
        case ITOF:
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:    
        case NEXT:
        case WNEXT:
        case PREV:
        case LEN:
        case STB:
        case MEM:
        case STRUCT:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case PEEK:
        case HELP:
        case SYSTEM:
        case TYPE:
                /*case iUNDEFINED:
                  case iREMEMBER:*/
        case ASCII:
        case FORMAT:
        case SIZE:
        case MAKELIST:
        case REFER:
        case DEREF:    
        case LOAD:
            node_lnk_1(e) = resolve_refs (node_lnk_1(e));
        default:
            return (e);
    }
}


/*void push_scope(name_list)
NODEPTR name_list;
{
    unsigned int level = node_env_level(EnvStack[EnvTop].env) + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = env_node(level, 0, NULL);
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}
*/

/*
void push_scope1(name_list,history)
NODEPTR name_list;
unsigned int history;
{
    unsigned int level = node_env_level(EnvStack[EnvTop].env) + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        if (history == 0)
            history=1;
        EnvStack[EnvTop].env = env_node1(level, 0, NULL, history);
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}
*/

void push_scope2(name_list,history,framed)
NODEPTR name_list;
unsigned int history;
unsigned framed;
{
    unsigned int level = node_env_level(EnvStack[EnvTop].env) + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        if (history == 0)
            history=1;
        EnvStack[EnvTop].env = env_node2(level, 0, NULL, history, framed);
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}

NODEPTR pop_scope()
{
    return (EnvStack[EnvTop--].env);
}

void install_name_list(name_list)
NODEPTR name_list;
{
    NODEPTR id;
    if (name_list == NULL)
        return;
    while (node_type(name_list) == COMMA) {
	id = node_lnk_1(name_list);
	install_local_id (node_sym_name(id));
	name_list = node_lnk_2(name_list);
    }
    install_local_id (node_sym_name(name_list));
}

NODEPTR lookup_env(s)
char *s;
{
    register NODEPTR np;
    register int level;

        /*printf("looking up %s\n",s);*/
    for (level = EnvTop; level >= EnvBase; level--)
        for (np = node_env_idlist(EnvStack[level].env);
	     np != NULL; np = node_next(np))
            if (strcmp (s, node_nm_name(np)) == 0)
                return (np);
    return (install_global_id (s));
}

NODEPTR install_global_id(name)
char *name;
{
    NODEPTR *np; 
    register    NODEPTR id;
    np = &(node_env_idlist(EnvStack[EnvBase].env));
    node_env_history(EnvStack[EnvBase].env)=1;
    node_env_framed(EnvStack[EnvBase].env)=0;
        /*printf("installing global: %s \n",name);*/
    if (DebugLevel>2) {
        printf("Global: history=%d\n",node_env_history(EnvStack[EnvBase].env));
        printf("Global: framed=%d\n",node_env_framed(EnvStack[EnvBase].env));
    }
    id = name_node ((unsigned)isupper ((char)*name), name, 0, offset2 (GlobalMemUsed++,node_env_history(EnvStack[EnvBase].env)),node_env_framed(EnvStack[EnvBase].env) );
    if (GlobalMemUsed >= GLOBALMEMSIZE)
        error_break0("global memory full");
    else
	new_loc2 (mem_loc (id), node_type(id), 1, 0);
    while (*np != NULL) {
        np = &(node_next(*np) );
    }
    return (*np = id);
}

void install_local_id(name)
char *name;
{
    NODEPTR *np; 
    register    NODEPTR id;
    np = &(node_env_idlist(EnvStack[EnvTop].env));
    if (DebugLevel>2) {
        printf("Local: name=%s\n",name);
        printf("Local: history=%d\n",node_env_history(EnvStack[EnvTop].env));
        printf("Local: framed=%d\n",node_env_framed(EnvStack[EnvTop].env));
    }
    id = name_node ((unsigned)isupper ((char)*name), name, node_env_level(EnvStack[EnvTop].env),
    		    offset2 (node_env_size(EnvStack[EnvTop].env)++, node_env_history(EnvStack[EnvTop].env)),node_env_framed(EnvStack[EnvTop].env) );
    while (*np != NULL) {
        np = &(node_next(*np) );
    }
    *np = id;
}

void syntax_error(n)
int n;
{
    yyerrorcount += 1;
    printf ("***Syntax Error");
    if (InputFile != stdin)
        printf (" on line %d of \"%s\"", yylineno-1, InputFilename);
    printf (": \n");
    fflush(stdout);
    syntax_msg (n);
    printf (".\n");
    fflush(stdout);
}

void syntax_msg(n)
int n;
{
    switch (n) {
            /*case N1:
            printf ("one identifier expected after quantifier");
            break;*/
        case MQS:
            printf ("missing quote in string");
            break;
        case MDF:
            printf ("missing $ in float");
            break;
        case EIF:
            printf ("error in float");
            break;    
        case STL:
            printf ("string too long");
            break;
        case IPTL:
            printf ("input line too long");
            break;
        case ERA:
            printf ("integer out of range");
            break;
        case UNK:
            printf ("unknow character");
            break;
/*        case FTL:
            printf ("float too long");
            break;*/
        case H1:
            printf ("history should be an integer greater than zero");
            break;     
        case IDN:
            printf ("syntax error, comment does not end properly");
            break;
/*        case NLC:
            printf ("syntax error, comma without a following identifier");
            break;*/
/*        case END:
            printf ("\"%s\" unexpected (command must be terminated by \".\")",
                                                                   InputText);
                                                                   break;*/
        default:
            break;
/*            printf ("\"%s\" unexpected", InputText);
            if (MINTOKEN <= n && n <= MAXTOKEN)
            printf (" (\"%s\" expected)", Keyword[n-MINTOKEN]);*/
    }
}

void yyerror(char const *s)
{
    int i;
    
    if (EOFILE==0) {
        yyerrorcount += 1;
        if (!InterruptSet) {
            printf("%s\n",Linetext);
            for (i=1; i<yylloc.first_column; i++)
                printf(" ");
            for (i=yylloc.first_column; i<=yylloc.last_column; i++)
                printf("^");
            newline;
            printf("***Syntax Error on line %d and column %d of \"%s\": `%s' unexpected\n\n",yylineno,yylloc.first_column,InputFilename,yytext);
        }
    }
}

